__all__ = ["AgeRating", "ComicInfo", "Manga", "Page", "PageType", "YesNo"]

from collections.abc import Callable
from datetime import date
from enum import Enum
from pathlib import Path
from typing import Optional

from natsort import humansorted, ns
from PIL import Image
from pydantic import HttpUrl, NonNegativeFloat
from pydantic_xml import attr, computed_attr, element, wrapped

from perdoo.metadata._base import PascalModel
from perdoo.settings import Naming


def str_to_list(value: str | None) -> list[str]:
    if not value:
        return []
    return humansorted({x.replace('"', "").strip() for x in value.split(",")}, alg=ns.NA | ns.G)


def list_to_str(value: list[str]) -> str | None:
    if not value:
        return None
    return ",".join(f'"{x}"' if "," in x else x for x in value)


class YesNo(Enum):
    UNKNOWN = "Unknown"
    NO = "No"
    YES = "Yes"

    @staticmethod
    def load(value: str) -> "YesNo":
        for entry in YesNo:
            if entry.value.replace(" ", "").casefold() == value.replace(" ", "").casefold():
                return entry
        raise ValueError(f"'{value}' isn't a valid YesNo")

    def __str__(self) -> str:
        return self.value


class Manga(Enum):
    UNKNOWN = "Unknown"
    NO = "No"
    YES = "Yes"
    YES_AND_RIGHT_TO_LEFT = "YesAndRightToLeft"

    @staticmethod
    def load(value: str) -> "Manga":
        for entry in Manga:
            if entry.value.replace(" ", "").casefold() == value.replace(" ", "").casefold():
                return entry
        raise ValueError(f"'{value}' isn't a valid Manga")

    def __str__(self) -> str:
        return self.value


class AgeRating(Enum):
    UNKNOWN = "Unknown"
    ADULTS_ONLY = "Adults Only 18+"
    EARLY_CHILDHOOD = "Early Childhood"
    EVERYONE = "Everyone"
    EVERYONE_10 = "Everyone 10+"
    G = "G"
    KIDS_TO_ADULTS = "Kids to Adults"
    M = "M"
    MA15 = "MA15+"
    MATURE_17 = "Mature 17+"
    PG = "PG"
    R18 = "R18+"
    RATING_PENDING = "Rating Pending"
    TEEN = "Teen"
    X18 = "X18+"

    @staticmethod
    def load(value: str) -> "AgeRating":
        for entry in AgeRating:
            if entry.value.replace(" ", "").casefold() == value.replace(" ", "").casefold():
                return entry
        raise ValueError(f"'{value}' isn't a valid AgeRating")

    def __str__(self) -> str:
        return self.value


class PageType(Enum):
    FRONT_COVER = "FrontCover"
    INNER_COVER = "InnerCover"
    ROUNDUP = "Roundup"
    STORY = "Story"
    ADVERTISEMENT = "Advertisement"
    EDITORIAL = "Editorial"
    LETTERS = "Letters"
    PREVIEW = "Preview"
    BACK_COVER = "BackCover"
    OTHER = "Other"
    DELETED = "Deleted"

    @staticmethod
    def load(value: str) -> "PageType":
        for entry in PageType:
            if entry.value.replace(" ", "").casefold() == value.replace(" ", "").casefold():
                return entry
        raise ValueError(f"'{value}' isn't a valid PageType")

    def __str__(self) -> str:
        return self.value


class Page(PascalModel):
    bookmark: str | None = attr(default=None)
    double_page: bool = attr(default=False)
    image: int = attr()
    image_height: int | None = attr(default=None)
    image_size: int = attr(default=0)
    image_width: int | None = attr(default=None)
    key: str | None = attr(default=None)
    type: PageType = attr(default=PageType.STORY)

    def __lt__(self, other) -> int:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.image < other.image

    def __eq__(self, other) -> bool:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.image == other.image

    def __hash__(self) -> int:
        return hash((type(self), self.image))

    @staticmethod
    def from_path(file: Path, index: int, is_final_page: bool, page: Optional["Page"]) -> "Page":
        if page:
            page_type = page.type
        elif index == 0:
            page_type = PageType.FRONT_COVER
        elif is_final_page:
            page_type = PageType.BACK_COVER
        else:
            page_type = PageType.STORY
        with Image.open(file) as img:
            width, height = img.size
        return Page(
            image=index,
            type=page_type,
            double_page=width >= height,
            image_size=file.stat().st_size,
            image_height=height,
            image_width=width,
        )


class ComicInfo(PascalModel):
    age_rating: AgeRating = element(default=AgeRating.UNKNOWN)
    alternate_count: int | None = element(default=None)
    alternate_number: str | None = element(default=None)
    alternate_series: str | None = element(default=None)
    black_and_white: YesNo = element(default=YesNo.UNKNOWN)
    characters: str | None = element(default=None)
    colorist: str | None = element(default=None)
    community_rating: NonNegativeFloat | None = element(default=None, le=5)
    count: int | None = element(default=None)
    cover_artist: str | None = element(default=None)
    day: int | None = element(default=None)
    editor: str | None = element(default=None)
    format: str | None = element(default=None)
    genre: str | None = element(default=None)
    imprint: str | None = element(default=None)
    inker: str | None = element(default=None)
    language_iso: str | None = element(tag="LanguageISO", default=None)
    letterer: str | None = element(default=None)
    locations: str | None = element(default=None)
    main_character_or_team: str | None = element(default=None)
    manga: Manga = element(default=Manga.UNKNOWN)
    month: int | None = element(default=None)
    notes: str | None = element(default=None)
    number: str | None = element(default=None)
    page_count: int = element(default=0)
    pages: list[Page] = wrapped(path="Pages", entity=element(tag="Page", default_factory=list))
    penciller: str | None = element(default=None)
    publisher: str | None = element(default=None)
    review: str | None = element(default=None)
    scan_information: str | None = element(default=None)
    series: str | None = element(default=None)
    series_group: str | None = element(default=None)
    story_arc: str | None = element(default=None)
    summary: str | None = element(default=None)
    teams: str | None = element(default=None)
    title: str | None = element(default=None)
    volume: int | None = element(default=None)
    web: HttpUrl | None = element(default=None)
    writer: str | None = element(default=None)
    year: int | None = element(default=None)

    @computed_attr(ns="xsi", name="noNamespaceSchemaLocation")
    def schema_location(self) -> str:
        return "https://raw.githubusercontent.com/Buried-In-Code/Schemas/main/schemas/v2.0/ComicInfo.xsd"

    @property
    def cover_date(self) -> date | None:
        if not self.year:
            return None
        return date(self.year, self.month or 1, self.day or 1)

    @cover_date.setter
    def cover_date(self, value: date | None) -> None:
        self.year = value.year if value else None
        self.month = value.month if value else None
        self.day = value.day if value else None

    @property
    def credits(self) -> dict[str, list[str]]:
        output = {}
        for role, attribute in (
            ("Writer", self.writer),
            ("Penciller", self.penciller),
            ("Inker", self.inker),
            ("Colorist", self.colorist),
            ("Letterer", self.letterer),
            ("Cover Artist", self.cover_artist),
            ("Editor", self.editor),
        ):
            if not attribute:
                continue
            creators = str_to_list(value=attribute)
            for creator in creators:
                output.setdefault(creator, []).append(role)
        return output

    @credits.setter
    def credits(self, value: dict[str, list[str]]) -> None:
        def get_creators(role: str) -> list[str]:
            return humansorted(
                {
                    creator
                    for creator, roles in value.items()
                    if any(role.casefold() == x.casefold() for x in roles)
                },
                alg=ns.NA | ns.G,
            )

        self.writer = list_to_str(value=get_creators(role="Writer"))
        self.penciller = list_to_str(value=get_creators(role="Penciller"))
        self.inker = list_to_str(value=get_creators(role="Inker"))
        self.colorist = list_to_str(value=get_creators(role="Colorist"))
        self.letterer = list_to_str(value=get_creators(role="Letterers"))
        self.cover_artist = list_to_str(value=get_creators(role="Cover Artist"))
        self.editor = list_to_str(value=get_creators(role="Editor"))

    @property
    def genre_list(self) -> list[str]:
        return str_to_list(value=self.genre)

    @genre_list.setter
    def genre_list(self, value: list[str]) -> None:
        self.genre = list_to_str(value=value)

    @property
    def character_list(self) -> list[str]:
        return str_to_list(value=self.characters)

    @character_list.setter
    def character_list(self, value: list[str]) -> None:
        self.characters = list_to_str(value=value)

    @property
    def team_list(self) -> list[str]:
        return str_to_list(value=self.teams)

    @team_list.setter
    def team_list(self, value: list[str]) -> None:
        self.teams = list_to_str(value=value)

    @property
    def location_list(self) -> list[str]:
        return str_to_list(value=self.locations)

    @location_list.setter
    def location_list(self, value: list[str]) -> None:
        self.locations = list_to_str(value=value)

    @property
    def story_arc_list(self) -> list[str]:
        return str_to_list(value=self.story_arc)

    @story_arc_list.setter
    def story_arc_list(self, value: list[str]) -> None:
        self.story_arc = list_to_str(value=value)

    def get_filename(self, settings: Naming) -> str:
        from perdoo.metadata.metron_info import Format

        return self.evaluate_pattern(
            pattern_map=PATTERN_MAP,
            pattern={
                Format.ANNUAL.value: settings.annual or settings.default,
                Format.DIGITAL_CHAPTER.value: settings.digital_chapter or settings.default,
                Format.GRAPHIC_NOVEL.value: settings.graphic_novel or settings.default,
                Format.HARDCOVER.value: settings.hardcover or settings.default,
                Format.LIMITED_SERIES.value: settings.limited_series or settings.default,
                Format.OMNIBUS.value: settings.omnibus or settings.default,
                Format.ONE_SHOT.value: settings.one_shot or settings.default,
                Format.SINGLE_ISSUE.value: settings.single_issue or settings.default,
                Format.TRADE_PAPERBACK.value: settings.trade_paperback or settings.default,
            }.get(self.format, settings.default),
        )


PATTERN_MAP: dict[str, Callable[[ComicInfo], str | int | None]] = {
    "cover-date": lambda x: x.cover_date,
    "cover-day": lambda x: x.day,
    "cover-month": lambda x: x.month,
    "cover-year": lambda x: x.year,
    "format": lambda x: x.format,
    "id": lambda _: None,
    "imprint": lambda x: x.imprint,
    "isbn": lambda _: None,
    "issue-count": lambda x: x.count,
    "lang": lambda x: x.language_iso,
    "number": lambda x: x.number,
    "publisher-id": lambda _: None,
    "publisher-name": lambda x: x.publisher,
    "series-id": lambda _: None,
    "series-name": lambda x: x.series,
    "series-sort-name": lambda _: None,
    "series-year": lambda x: x.volume if x.volume > 1900 else None,
    "store-date": lambda _: None,
    "store-day": lambda _: None,
    "store-month": lambda _: None,
    "store-year": lambda _: None,
    "title": lambda x: x.title,
    "upc": lambda _: None,
    "volume": lambda x: x.volume if x.volume < 1900 else None,
}



__all__ = ["CBZArchive"]

import logging
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional
from zipfile import ZIP_DEFLATED, BadZipFile, ZipFile

from perdoo.archives._base import BaseArchive
from perdoo.archives.zipfile_remove import ZipFileRemove
from perdoo.utils import list_files

LOGGER = logging.getLogger(__name__)


class CBZArchive(BaseArchive):
    def list_filenames(self) -> list[str]:
        try:
            with ZipFile(self.path, "r") as stream:
                return stream.namelist()
        except BadZipFile:
            LOGGER.exception("Unable to read %s", self.path.name)
            return []

    def read_file(self, filename: str) -> bytes:
        try:
            with ZipFile(self.path, "r") as zip_file, zip_file.open(filename) as file:
                return file.read()
        except (BadZipFile, KeyError):
            LOGGER.exception("Unable to read %s", self.path.name)
            return b""

    def remove_file(self, filename: str) -> bool:
        if filename not in self.list_filenames():
            return True
        try:
            with ZipFileRemove(self.path, "a") as stream:
                stream.remove(filename)
            return True
        except BadZipFile:
            LOGGER.exception("")
            return False

    def write_file(self, filename: str, data: str) -> bool:
        try:
            with ZipFileRemove(self.path, "a") as stream:
                if filename in stream.namelist():
                    stream.remove(filename)
                stream.writestr(filename, data)
            return True
        except BadZipFile:
            LOGGER.exception("")
            return False

    def extract_files(self, destination: Path) -> bool:
        try:
            with ZipFile(self.path, "r") as stream:
                stream.extractall(path=destination)
            return True
        except BadZipFile:
            LOGGER.exception("")
            return False

    @classmethod
    def archive_files(cls, src: Path, output_name: str, files: list[Path]) -> Path | None:
        output_file = src.parent / f"{output_name}.cbz"
        try:
            with ZipFile(output_file, "w", ZIP_DEFLATED) as stream:
                for file in files:
                    stream.write(file, arcname=file.name)
            return output_file
        except BadZipFile:
            LOGGER.exception("")
            return None

    @staticmethod
    def convert(old_archive: BaseArchive) -> Optional["CBZArchive"]:
        with TemporaryDirectory(prefix=f"{old_archive.path.stem}_") as temp_str:
            temp_folder = Path(temp_str)
            if not old_archive.extract_files(destination=temp_folder):
                return None
            archive_file = CBZArchive.archive_files(
                src=temp_folder, output_name=old_archive.path.stem, files=list_files(temp_folder)
            )
            if not archive_file:
                return None
            new_file = old_archive.path.with_suffix(".cbz")
            old_archive.path.unlink(missing_ok=True)
            shutil.move(archive_file, new_file)
            return CBZArchive(path=new_file)



__all__ = ["BaseService"]

from abc import abstractmethod
from typing import Generic, TypeVar

from perdoo.metadata import ComicInfo, MetronInfo
from perdoo.utils import IssueSearch, Search, SeriesSearch

S = TypeVar("S")
C = TypeVar("C")


class BaseService(Generic[S, C]):
    @abstractmethod
    def _search_series(
        self, name: str | None, volume: int | None, year: int | None
    ) -> int | None: ...

    @abstractmethod
    def fetch_series(self, search: SeriesSearch) -> S | None: ...

    @abstractmethod
    def _search_issue(self, series_id: int, number: str | None) -> int | None: ...

    @abstractmethod
    def fetch_issue(self, series_id: int, search: IssueSearch) -> C | None: ...

    @abstractmethod
    def _process_metron_info(self, series: S, issue: C) -> MetronInfo | None: ...

    @abstractmethod
    def _process_comic_info(self, series: S, issue: C) -> ComicInfo | None: ...

    @abstractmethod
    def fetch(self, search: Search) -> tuple[MetronInfo | None, ComicInfo | None]: ...



__all__ = [
    "GTIN",
    "AgeRating",
    "AlternativeName",
    "Arc",
    "Credit",
    "Format",
    "Id",
    "InformationSource",
    "MetronInfo",
    "Price",
    "Publisher",
    "Resource",
    "Role",
    "Series",
    "Universe",
    "Url",
]

from collections.abc import Callable
from datetime import date, datetime
from decimal import Decimal
from enum import Enum
from typing import Generic, TypeVar

from pydantic import HttpUrl, NonNegativeInt, PositiveInt
from pydantic_xml import attr, computed_attr, element, wrapped

from perdoo.metadata._base import PascalModel
from perdoo.settings import Naming

T = TypeVar("T")


class AgeRating(Enum):
    UNKNOWN = "Unknown"
    EVERYONE = "Everyone"
    TEEN = "Teen"
    TEEN_PLUS = "Teen Plus"
    MATURE = "Mature"
    EXPLICIT = "Explicit"
    ADULT = "Adult"

    @staticmethod
    def load(value: str) -> "AgeRating":
        for entry in AgeRating:
            if entry.value.replace(" ", "").casefold() == value.replace(" ", "").casefold():
                return entry
        raise ValueError(f"'{value}' isn't a valid AgeRating")

    def __str__(self) -> str:
        return self.value


class Arc(PascalModel):
    id: str | None = attr(name="id", default=None)
    name: str = element()
    number: PositiveInt | None = element(default=None)

    def __lt__(self, other) -> int:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.name < other.name

    def __eq__(self, other) -> bool:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.name == other.name

    def __hash__(self) -> int:
        return hash((type(self), self.name))


class Resource(PascalModel, Generic[T]):
    value: T
    id: str | None = attr(name="id", default=None)

    def __lt__(self, other) -> int:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value < other.value

    def __eq__(self, other) -> bool:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value == other.value

    def __hash__(self) -> int:
        return hash((type(self), self.value))


class Role(Enum):
    WRITER = "Writer"
    SCRIPT = "Script"
    STORY = "Story"
    PLOT = "Plot"
    INTERVIEWER = "Interviewer"
    ARTIST = "Artist"
    PENCILLER = "Penciller"
    BREAKDOWNS = "Breakdowns"
    ILLUSTRATOR = "Illustrator"
    LAYOUTS = "Layouts"
    INKER = "Inker"
    EMBELLISHER = "Embellisher"
    FINISHES = "Finishes"
    INK_ASSISTS = "Ink Assists"
    COLORIST = "Colorist"
    COLOR_SEPARATIONS = "Color Separations"
    COLOR_ASSISTS = "Color Assists"
    COLOR_FLATS = "Color Flats"
    DIGITAL_ART_TECHNICIAN = "Digital Art Technician"
    GRAY_TONE = "Gray Tone"
    LETTERER = "Letterer"
    COVER = "Cover"
    EDITOR = "Editor"
    CONSULTING_EDITOR = "Consulting Editor"
    ASSISTANT_EDITOR = "Assistant Editor"
    ASSOCIATE_EDITOR = "Associate Editor"
    GROUP_EDITOR = "Group Editor"
    SENIOR_EDITOR = "Senior Editor"
    MANAGING_EDITOR = "Managing Editor"
    COLLECTION_EDITOR = "Collection Editor"
    PRODUCTION = "Production"
    DESIGNER = "Designer"
    LOGO_DESIGN = "Logo Design"
    TRANSLATOR = "Translator"
    SUPERVISING_EDITOR = "Supervising Editor"
    EXECUTIVE_EDITOR = "Executive Editor"
    EDITOR_IN_CHIEF = "Editor In Chief"
    PRESIDENT = "President"
    PUBLISHER = "Publisher"
    CHIEF_CREATIVE_OFFICER = "Chief Creative Officer"
    EXECUTIVE_PRODUCER = "Executive Producer"
    OTHER = "Other"

    @staticmethod
    def load(value: str) -> "Role":
        for entry in Role:
            if entry.value.replace(" ", "").casefold() == value.replace(" ", "").casefold():
                return entry
        raise ValueError(f"'{value}' isn't a valid Role")

    def __str__(self) -> str:
        return self.value


class Credit(PascalModel):
    creator: Resource[str] = element()
    roles: list[Resource[Role]] = wrapped(
        path="Roles", entity=element(tag="Role", default_factory=list)
    )

    def __lt__(self, other) -> int:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.creator < other.creator

    def __eq__(self, other) -> bool:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.creator == other.creator

    def __hash__(self) -> int:
        return hash((type(self), self.creator))


class GTIN(PascalModel):
    isbn: str | None = element(tag="ISBN", default=None)
    upc: str | None = element(tag="UPC", default=None)


class InformationSource(Enum):
    ANILIST = "AniList"
    COMIC_VINE = "Comic Vine"
    GRAND_COMICS_DATABASE = "Grand Comics Database"
    KITSU = "Kitsu"
    MANGA_DEX = "MangaDex"
    MANGA_UPDATES = "MangaUpdates"
    MARVEL = "Marvel"
    METRON = "Metron"
    MYANIMELIST = "MyAnimeList"
    LEAGUE_OF_COMIC_GEEKS = "League of Comic Geeks"

    @staticmethod
    def load(value: str) -> "InformationSource":
        for entry in InformationSource:
            if entry.value.replace(" ", "").casefold() == value.replace(" ", "").casefold():
                return entry
        raise ValueError(f"'{value}' isn't a valid InformationSource")

    def __str__(self) -> str:
        return self.value


class Id(PascalModel):
    primary: bool = attr(name="primary", default=False)
    source: InformationSource = attr(name="source")
    value: str

    def __lt__(self, other) -> int:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.source < other.source

    def __eq__(self, other) -> bool:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.source == other.source

    def __hash__(self) -> int:
        return hash((type(self), self.source))


class Price(PascalModel):
    country: str = attr(name="country")
    value: Decimal

    def __lt__(self, other) -> int:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.country < other.country

    def __eq__(self, other) -> bool:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.country == other.country

    def __hash__(self) -> int:
        return hash((type(self), self.country))


class Publisher(PascalModel):
    id: str | None = attr(name="id", default=None)
    imprint: Resource[str] | None = element(default=None)
    name: str = element()


class AlternativeName(Resource[str]):
    lang: str = attr(name="lang", default="en")


class Format(Enum):
    ANNUAL = "Annual"
    DIGITAL_CHAPTER = "Digital Chapter"
    GRAPHIC_NOVEL = "Graphic Novel"
    HARDCOVER = "Hardcover"
    LIMITED_SERIES = "Limited Series"
    OMNIBUS = "Omnibus"
    ONE_SHOT = "One-Shot"
    SINGLE_ISSUE = "Single Issue"
    TRADE_PAPERBACK = "Trade Paperback"

    @staticmethod
    def load(value: str) -> "Format":
        for entry in Format:
            if entry.value.replace(" ", "").casefold() == value.replace(" ", "").casefold():
                return entry
        raise ValueError(f"'{value}' isn't a valid Format")

    def __str__(self) -> str:
        return self.value


class Series(PascalModel):
    alternative_names: list[AlternativeName] = wrapped(
        path="AlternativeNames", entity=element(tag="AlternativeName", default_factory=list)
    )
    format: Format | None = element(default=None)
    id: str | None = attr(name="id", default=None)
    issue_count: PositiveInt | None = element(default=None)
    lang: str = attr(name="lang", default="en")
    name: str = element()
    sort_name: str | None = element(default=None)
    start_year: int | None = element(default=None)
    volume: NonNegativeInt | None = element(default=None)
    volume_count: PositiveInt | None = element(default=None)


class Universe(PascalModel):
    designation: str | None = element(default=None)
    id: str | None = attr(name="id", default=None)
    name: str = element()

    def __lt__(self, other) -> int:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.name < other.name

    def __eq__(self, other) -> bool:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.name == other.name

    def __hash__(self) -> int:
        return hash((type(self), self.name))


class Url(PascalModel):
    primary: bool = attr(name="primary", default=False)
    value: HttpUrl

    def __lt__(self, other) -> int:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value < other.value

    def __eq__(self, other) -> bool:  # noqa: ANN001
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.value == other.value

    def __hash__(self) -> int:
        return hash((type(self), self.value))


class MetronInfo(PascalModel):
    age_rating: AgeRating = element(default=AgeRating.UNKNOWN)
    arcs: list[Arc] = wrapped(path="Arcs", entity=element(tag="Arc", default_factory=list))
    characters: list[Resource[str]] = wrapped(
        path="Characters", entity=element(tag="Character", default_factory=list)
    )
    collection_title: str | None = element(default=None)
    cover_date: date | None = element(default=None)
    credits: list[Credit] = wrapped(
        path="Credits", entity=element(tag="Credit", default_factory=list)
    )
    genres: list[Resource[str]] = wrapped(
        path="Genres", entity=element(tag="Genre", default_factory=list)
    )
    gtin: GTIN | None = element(tag="GTIN", default=None)
    ids: list[Id] = wrapped("IDS", entity=element(tag="ID", default_factory=list))
    last_modified: datetime | None = element(default=None)
    locations: list[Resource[str]] = wrapped(
        path="Locations", entity=element(tag="Location", default_factory=list)
    )
    manga_volume: str | None = element(default=None)
    notes: str | None = element(default=None)
    number: str | None = element(default=None)
    page_count: NonNegativeInt = element(default=0)
    prices: list[Price] = wrapped(path="Prices", entity=element(tag="Price", default_factory=list))
    publisher: Publisher | None = element(default=None)
    reprints: list[Resource[str]] = wrapped(
        path="Reprints", entity=element(tag="Reprint", default_factory=list)
    )
    series: Series = element()
    store_date: date | None = element(default=None)
    stories: list[Resource[str]] = wrapped(
        path="Stories", entity=element(tag="Story", default_factory=list)
    )
    summary: str | None = element(default=None)
    tags: list[Resource[str]] = wrapped(
        path="Tags", entity=element(tag="Tag", default_factory=list)
    )
    teams: list[Resource[str]] = wrapped(
        path="Teams", entity=element(tag="Team", default_factory=list)
    )
    universes: list[Universe] = wrapped(
        path="Universes", entity=element(tag="Universe", default_factory=list)
    )
    urls: list[Url] = wrapped(path="URLS", entity=element(tag="URLs", default_factory=list))

    @computed_attr(ns="xsi", name="noNamespaceSchemaLocation")
    def schema_location(self) -> str:
        return "https://raw.githubusercontent.com/Metron-Project/metroninfo/master/schema/v1.0/MetronInfo.xsd"

    def get_filename(self, settings: Naming) -> str:
        return self.evaluate_pattern(
            pattern_map=PATTERN_MAP,
            pattern={
                Format.ANNUAL: settings.annual or settings.default,
                Format.DIGITAL_CHAPTER: settings.digital_chapter or settings.default,
                Format.GRAPHIC_NOVEL: settings.graphic_novel or settings.default,
                Format.HARDCOVER: settings.hardcover or settings.default,
                Format.LIMITED_SERIES: settings.limited_series or settings.default,
                Format.OMNIBUS: settings.omnibus or settings.default,
                Format.ONE_SHOT: settings.one_shot or settings.default,
                Format.SINGLE_ISSUE: settings.single_issue or settings.default,
                Format.TRADE_PAPERBACK: settings.trade_paperback or settings.default,
            }.get(self.series.format, settings.default),
        )


PATTERN_MAP: dict[str, Callable[[MetronInfo], str | int | None]] = {
    "cover-date": lambda x: x.cover_date,
    "cover-day": lambda x: x.cover_date.day if x.cover_date else None,
    "cover-month": lambda x: x.cover_date.month if x.cover_date else None,
    "cover-year": lambda x: x.cover_date.year if x.cover_date else None,
    "format": lambda x: x.series.format.value if x.series.format else None,
    "id": lambda x: next(iter(i.value for i in x.ids if i.primary), None),
    "imprint": lambda x: x.publisher.imprint.value if x.publisher and x.publisher.imprint else None,
    "isbn": lambda x: x.gtin.isbn if x.gtin else None,
    "issue-count": lambda x: x.series.issue_count,
    "lang": lambda x: x.series.lang,
    "number": lambda x: x.number,
    "publisher-id": lambda x: x.publisher.id if x.publisher else None,
    "publisher-name": lambda x: x.publisher.name if x.publisher else None,
    "series-id": lambda x: x.series.id,
    "series-name": lambda x: x.series.name,
    "series-sort-name": lambda x: x.series.sort_name,
    "series-year": lambda x: x.series.start_year,
    "store-date": lambda x: x.store_date or "",
    "store-year": lambda x: x.store_date.year if x.store_date else "",
    "store-month": lambda x: x.store_date.month if x.store_date else "",
    "store-day": lambda x: x.store_date.day if x.store_date else "",
    "title": lambda x: x.collection_title,
    "upc": lambda x: x.gtin.upc if x.gtin else None,
    "volume": lambda x: x.series.volume,
}



__all__ = ["Metron"]

import logging
from datetime import datetime

from mokkari.exceptions import ApiError
from mokkari.schemas.issue import Issue
from mokkari.schemas.series import Series
from mokkari.session import Session as Mokkari
from mokkari.sqlite_cache import SqliteCache
from natsort import humansorted, ns
from rich.prompt import Confirm, Prompt

from perdoo import get_cache_root
from perdoo.console import CONSOLE, create_menu
from perdoo.metadata import ComicInfo, MetronInfo
from perdoo.metadata.metron_info import InformationSource
from perdoo.services._base import BaseService
from perdoo.settings import Metron as MetronSettings
from perdoo.utils import IssueSearch, Search, SeriesSearch

LOGGER = logging.getLogger(__name__)


class Metron(BaseService[Series, Issue]):
    def __init__(self, settings: MetronSettings):
        cache = SqliteCache(db_name=str(get_cache_root() / "mokkari.sqlite"), expire=14)
        self.session = Mokkari(username=settings.username, passwd=settings.password, cache=cache)

    def _search_series_by_comicvine(self, comicvine_id: int | None) -> int | None:
        if not comicvine_id:
            return None
        try:
            series = self.session.series_list({"cv_id": comicvine_id})
            if series and len(series) >= 1:
                return series[0].id
            return None
        except ApiError:
            LOGGER.exception("")
            return None

    def _search_series(self, name: str | None, volume: int | None, year: int | None) -> int | None:
        name = name or Prompt.ask("Series Name", console=CONSOLE)
        try:
            params = {"name": name}
            if volume:
                params["volume"] = volume
            if year:
                params["year_began"] = year
            options = sorted(
                self.session.series_list(params=params), key=lambda x: (x.display_name, x.volume)
            )
            if not options:
                LOGGER.warning(
                    "Unable to find any Series with the Name, Volume and YearBegan: '%s %s %s'",
                    name,
                    volume,
                    year,
                )
            search = name
            if volume:
                search += f" v{volume}"
            if year:
                search += f" ({year})"
            index = create_menu(
                options=[
                    f"{x.id} | {x.display_name} v{x.volume}"
                    if x.volume > 1
                    else f"{x.id} | {x.display_name}"
                    for x in options
                ],
                title="Metron Series",
                subtitle=f"Searching for Series '{search}'",
                default="None of the Above",
            )
            if index != 0:
                return options[index - 1].id
            if year:
                LOGGER.info("Searching again without the YearBegan")
                return self._search_series(name=name, volume=volume, year=None)
            if volume:
                LOGGER.info("Searching again without the Volume")
                return self._search_series(name=name, volume=None, year=None)
            if Confirm.ask("Search Again", console=CONSOLE):
                return self._search_series(name=None, volume=None, year=None)
            return None
        except ApiError:
            LOGGER.exception("")
            return None

    def fetch_series(self, search: SeriesSearch) -> Series | None:
        series_id = (
            search.metron
            or self._search_series_by_comicvine(comicvine_id=search.comicvine)
            or self._search_series(name=search.name, volume=search.volume, year=search.year)
        )
        if not series_id:
            return None
        try:
            series = self.session.series(_id=series_id)
            search.metron = series_id
            return series
        except ApiError:
            LOGGER.exception("")
            return None

    def _search_issue_by_comicvine(self, comicvine_id: int | None) -> int | None:
        if not comicvine_id:
            return None
        try:
            issues = self.session.issues_list({"cv_id": comicvine_id})
            if issues and len(issues) >= 1:
                return issues[0].id
            return None
        except ApiError:
            LOGGER.exception("")
            return None

    def _search_issue(self, series_id: int, number: str | None) -> int | None:
        try:
            options = humansorted(
                self.session.issues_list(
                    params={"series_id": series_id, "number": number}
                    if number
                    else {"series_id": series_id}
                ),
                key=lambda x: (x.number, x.issue_name),
                alg=ns.NA | ns.G,
            )
            if not options:
                LOGGER.warning(
                    "Unable to find any Issues with the SeriesId and Number: '%s %s'",
                    series_id,
                    number,
                )
            index = create_menu(
                options=[f"{x.id} | {x.issue_name}" for x in options],
                title="Metron Issue",
                subtitle=f"Searching for Issue #{number}" if number else "",
                default="None of the Above",
            )
            if index != 0:
                return options[index - 1].id
            if number:
                LOGGER.info("Searching again without the Number")
                return self._search_issue(series_id=series_id, number=None)
            return None
        except ApiError:
            LOGGER.exception("")
            return None

    def fetch_issue(self, series_id: int, search: IssueSearch) -> Issue | None:
        issue_id = (
            search.metron
            or self._search_issue_by_comicvine(comicvine_id=search.comicvine)
            or self._search_issue(series_id=series_id, number=search.number)
        )
        if not issue_id:
            return None
        try:
            issue = self.session.issue(_id=issue_id)
            search.metron = issue_id
            return issue
        except ApiError:
            LOGGER.exception("")
            return None

    def _process_metron_info(self, series: Series, issue: Issue) -> MetronInfo | None:
        from perdoo.metadata.metron_info import (
            GTIN,
            AgeRating,
            Arc,
            Credit,
            Format,
            Id,
            Price,
            Publisher,
            Resource,
            Role,
            Series,
            Universe,
            Url,
        )

        def load_role(value: str) -> Role:
            try:
                return Role.load(value=value.strip())
            except ValueError:
                return Role.OTHER

        ids = [Id(primary=True, source=InformationSource.METRON, value=str(issue.id))]
        if issue.cv_id:
            ids.append(Id(source=InformationSource.COMIC_VINE, value=str(issue.cv_id)))
        if issue.gcd_id:
            ids.append(Id(source=InformationSource.GRAND_COMICS_DATABASE, value=str(issue.gcd_id)))
        return MetronInfo(
            ids=ids,
            publisher=Publisher(
                id=str(series.publisher.id),
                name=series.publisher.name,
                imprint=Resource[str](id=str(series.imprint.id), value=series.imprint.name)
                if series.imprint
                else None,
            ),
            series=Series(
                id=str(series.id),
                name=series.name,
                sort_name=series.sort_name,
                volume=series.volume,
                format=Format.load(value=series.series_type.name),
                start_year=series.year_began,
            ),
            collection_title=issue.collection_title or None,
            number=issue.number,
            stories=[Resource[str](value=x) for x in issue.story_titles],
            summary=issue.desc,
            prices=[Price(country="US", value=issue.price)] if issue.price else [],
            cover_date=issue.cover_date,
            store_date=issue.store_date,
            page_count=issue.page_count or 0,
            genres=[Resource[str](id=str(x.id), value=x.name) for x in issue.series.genres],
            arcs=[Arc(id=str(x.id), name=x.name) for x in issue.arcs],
            characters=[Resource[str](id=str(x.id), value=x.name) for x in issue.characters],
            teams=[Resource[str](id=str(x.id), value=x.name) for x in issue.teams],
            universes=[Universe(id=str(x.id), name=x.name) for x in issue.universes],
            gtin=GTIN(isbn=issue.isbn or None, upc=issue.upc or None)
            if issue.isbn or issue.upc
            else None,
            age_rating=AgeRating.load(value=issue.rating.name),
            reprints=[Resource[str](id=str(x.id), value=x.issue) for x in issue.reprints],
            urls=[Url(primary=True, value=issue.resource_url)],
            credits=[
                Credit(
                    creator=Resource[str](id=str(x.id), value=x.creator),
                    roles=[
                        Resource[Role](id=str(r.id), value=load_role(value=r.name)) for r in x.role
                    ],
                )
                for x in issue.credits
            ],
            last_modified=datetime.now(),
        )

    def _process_comic_info(self, series: Series, issue: Issue) -> ComicInfo | None:
        from perdoo.metadata.comic_info import AgeRating

        def load_age_rating(value: str) -> AgeRating:
            try:
                return AgeRating.load(value=value.strip())
            except ValueError:
                return AgeRating.UNKNOWN

        comic_info = ComicInfo(
            title=issue.collection_title,
            series=series.name,
            number=issue.number,
            volume=series.volume,
            summary=issue.desc,
            publisher=series.publisher.name,
            web=issue.resource_url,
            page_count=issue.page_count or 0,
            format=series.series_type.name,
            age_rating=load_age_rating(value=issue.rating.name),
        )

        comic_info.cover_date = issue.cover_date
        comic_info.credits = {x.creator: [r.name for r in x.role] for x in issue.credits}
        comic_info.genre_list = [x.name for x in series.genres]
        comic_info.character_list = [x.name for x in issue.characters]
        comic_info.team_list = [x.name for x in issue.teams]
        comic_info.story_arc_list = [x.name for x in issue.arcs]

        return comic_info

    def fetch(self, search: Search) -> tuple[MetronInfo | None, ComicInfo | None]:
        if not search.series.metron and search.issue.metron:
            try:
                temp = self.session.issue(_id=search.issue.metron)
                search.series.metron = temp.series.id
            except ApiError:
                pass

        series = self.fetch_series(search=search.series)
        if not series:
            return None, None

        issue = self.fetch_issue(series_id=series.id, search=search.issue)
        if not issue:
            return None, None

        metron_info = self._process_metron_info(series=series, issue=issue)
        comic_info = self._process_comic_info(series=series, issue=issue)

        return metron_info, comic_info



__all__ = ["CB7Archive"]

import logging
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

from perdoo.archives._base import BaseArchive
from perdoo.utils import list_files

try:
    from py7zr import Bad7zFile, SevenZipFile

    py7zr_loaded = True
except ImportError:
    py7zr_loaded = False

LOGGER = logging.getLogger(__name__)


class CB7Archive(BaseArchive):
    def __init__(self, path: Path):
        if not py7zr_loaded:
            raise ImportError("Install Perdoo with the cb7 dependency group to use CB7 files.")
        super().__init__(path=path)

    def list_filenames(self) -> list[str]:
        try:
            with SevenZipFile(self.path, "r") as archive:
                return archive.getnames()
        except Bad7zFile:
            LOGGER.exception("Unable to read %s", self.path.name)
            return []

    def read_file(self, filename: str) -> bytes:
        try:
            with SevenZipFile(self.path, "r") as archive, archive.open(filename) as file:
                return file.read()
        except (Bad7zFile, KeyError):
            LOGGER.exception("Unable to read %s", self.path.name)
            return b""

    def extract_files(self, destination: Path) -> bool:
        try:
            with SevenZipFile(self.path, "r") as archive:
                archive.extractall(path=destination)
            return True
        except Bad7zFile:
            LOGGER.exception("")
            return False

    @classmethod
    def archive_files(cls, src: Path, output_name: str, files: list[Path]) -> Path | None:
        if not py7zr_loaded:
            raise ImportError("Install Perdoo with the cb7 dependency group to use CB7 files.")

        output_file = src.parent / f"{output_name}.cb7"
        try:
            with SevenZipFile(output_file, "w") as archive:
                for file in files:
                    archive.write(file, arcname=file.name)
            return output_file
        except Bad7zFile:
            LOGGER.exception("")
            return None

    @staticmethod
    def convert(old_archive: BaseArchive) -> Optional["CB7Archive"]:
        if not py7zr_loaded:
            raise ImportError("Install Perdoo with the cb7 dependency group to use CB7 files.")

        with TemporaryDirectory(prefix=f"{old_archive.path.stem}_") as temp_str:
            temp_folder = Path(temp_str)
            if not old_archive.extract_files(destination=temp_folder):
                return None
            archive_file = CB7Archive.archive_files(
                src=temp_folder, output_name=old_archive.path.stem, files=list_files(temp_folder)
            )
            if not archive_file:
                return None
            new_file = old_archive.path.with_suffix(".cb7")
            old_archive.path.unlink(missing_ok=True)
            shutil.move(archive_file, new_file)
            return CB7Archive(path=new_file)



__all__ = ["CONSOLE", "create_menu"]

import logging

from rich.console import Console
from rich.panel import Panel
from rich.prompt import IntPrompt
from rich.theme import Theme

LOGGER = logging.getLogger(__name__)

CONSOLE = Console(
    theme=Theme(
        {
            "prompt": "cyan",
            "prompt.border": "dim bright_cyan",
            "prompt.choices": "white",
            "prompt.default": "italic white",
            "title": "bold not dim blue",
            "subtitle": "not dim blue",
            "logging.level.debug": "dim white",
            "logging.level.info": "white",
            "logging.level.warning": "yellow",
            "logging.level.error": "red",
            "logging.level.critical": "bold bright_red",
        }
    )
)


def create_menu(
    options: list[str],
    title: str | None = None,
    subtitle: str | None = None,
    prompt: str = "Select",
    default: str | None = None,
) -> int | None:
    if not options:
        return 0
    panel_text = []
    padding = len(str(len(options)))
    for index, item in enumerate(options):
        panel_text.append(
            f"[markdown.item.number]{index + 1:>{padding}}.[/] [prompt.choices]{item}[/]"
        )
    if default:
        panel_text.append(f"[markdown.item.number]{0:>{padding}}.[/] [prompt.default]{default}[/]")
    CONSOLE.print(
        Panel(
            "\n".join(panel_text),
            title=f"[title]{title}[/]" if title else None,
            subtitle=f"[subtitle]{subtitle}[/]" if subtitle else None,
            border_style="prompt.border",
        )
    )
    selected = IntPrompt.ask(prompt=prompt, default=0 if default else None, console=CONSOLE)
    if (
        selected is None
        or selected < 0
        or selected > len(options)
        or (selected == 0 and not default)
    ):
        LOGGER.warning("Invalid Option: %s", selected)
        return create_menu(
            options=options, title=title, subtitle=subtitle, prompt=prompt, default=default
        )
    return selected



from perdoo.metadata._base import sanitize
from perdoo.metadata.comic_info import ComicInfo
from perdoo.metadata.metron_info import Format, MetronInfo, Publisher, Series
from perdoo.settings import Naming


def test_sanitize() -> None:
    assert sanitize("Example Title!") == "Example-Title!"
    assert sanitize("Example/Title: 123") == "ExampleTitle-123"
    assert sanitize("!@#$%^&*()[]{};':,.<>?/") == "!&"
    assert sanitize(None) is None


def test_metron_info_default_naming() -> None:
    obj = MetronInfo(
        publisher=Publisher(name="Example Publisher"),
        series=Series(name="Example Series", volume=1, format=Format.TRADE_PAPERBACK),
        number=2,
    )
    assert (
        obj.get_filename(settings=Naming())
        == "Example-Publisher/Example-Series-v1/Example-Series-v1_TPB_#02"
    )


def test_comic_info_default_naming() -> None:
    obj = ComicInfo(
        publisher="Example Publisher",
        series="Example Series",
        format="Trade Paperback",
        volume=1,
        number=2,
    )
    assert (
        obj.get_filename(settings=Naming())
        == "Example-Publisher/Example-Series-v1/Example-Series-v1_TPB_#02"
    )



__all__ = ["CBRArchive"]

import logging
from pathlib import Path
from typing import Optional

from rarfile import RarExecError, RarFile

from perdoo.archives._base import BaseArchive

LOGGER = logging.getLogger(__name__)


class CBRArchive(BaseArchive):
    def list_filenames(self) -> list[str]:
        try:
            with RarFile(self.path) as stream:
                return stream.namelist()
        except RarExecError:
            LOGGER.exception("Unable to read %s", self.path.name)
            return []

    def read_file(self, filename: str) -> bytes:
        try:
            with RarFile(self.path) as stream:
                return stream.read(filename)
        except RarExecError:
            LOGGER.exception("Unable to read %s", self.path.name)
            return b""

    def remove_file(self, filename: str) -> bool:
        raise NotImplementedError("Unable to remove a file from a CBR archive.")

    def write_file(self, filename: str, data: str) -> bool:
        raise NotImplementedError("Unable to write a file to a CBR archive.")

    def extract_files(self, destination: Path) -> bool:
        try:
            with RarFile(self.path, "r") as stream:
                stream.extractall(path=destination)
            return True
        except RarExecError:
            LOGGER.exception("")
            return False

    @classmethod
    def archive_files(cls, src: Path, output_name: str, files: list[Path]) -> Path | None:
        raise NotImplementedError("Unable to create a CBR archive.")

    @staticmethod
    def convert(old_archive: BaseArchive) -> Optional["CBRArchive"]:
        raise NotImplementedError("Unable to convert an archive to CBR.")



__all__ = ["app"]

from argparse import SUPPRESS
from typing import Annotated

from typer import Argument, Option, Typer

from perdoo.console import CONSOLE
from perdoo.settings import Settings
from perdoo.utils import flatten_dict

app = Typer(help="Commands for managing and configuring application settings.")


@app.command(name="view", help="Display the current and default settings.")
def view() -> None:
    settings = Settings.load()
    settings.display()


@app.command(name="locate", help="Display the path to the settings file.")
def locate() -> None:
    CONSOLE.print(Settings._file)  # noqa: SLF001


@app.command(name="update", help="Update the settings.")
def update(
    key: Annotated[str | None, Argument(show_default=False, help="The setting to update.")] = None,
    value: Annotated[
        str | None, Argument(show_default=False, help="The value to update the setting to.")
    ] = SUPPRESS,
    reset: Annotated[
        bool,
        Option(
            "--reset",
            help="Reset the specified setting to its default value. If no key is provided, reset all settings.",  # noqa: E501
        ),
    ] = False,
) -> None:
    if reset:
        Settings().save()
        CONSOLE.print("Settings reset")
    elif key:
        settings = Settings.load()
        if reset:
            settings_dict = flatten_dict(content=Settings().model_dump())
            if key in settings_dict:
                settings.update(key=key, value=settings_dict[key])
                settings.save()
                CONSOLE.print(f"'{key}' Reset")
            else:
                CONSOLE.print(f"No Config key: '{key}'", style="logging.level.critical")
        elif value is not SUPPRESS:
            settings.update(key=key, value=value)
            settings.save()
            CONSOLE.print(f"Updated '{key}' to {value}")



__all__ = [
    "IssueSearch",
    "Search",
    "SeriesSearch",
    "blank_is_none",
    "delete_empty_folders",
    "flatten_dict",
    "list_files",
]

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from natsort import humansorted, ns

LOGGER = logging.getLogger(__name__)


@dataclass
class SeriesSearch:
    name: str
    volume: int | None = None
    year: int | None = None
    comicvine: int | None = None
    marvel: int | None = None
    metron: int | None = None


@dataclass
class IssueSearch:
    number: str | None = None
    comicvine: int | None = None
    marvel: int | None = None
    metron: int | None = None


@dataclass
class Search:
    series: SeriesSearch
    issue: IssueSearch


def list_files(path: Path, *extensions: str) -> list[Path]:
    files = []
    for file in path.iterdir():
        if file.is_file():
            if not file.name.startswith(".") and (
                not extensions or file.suffix.lower() in extensions
            ):
                files.append(file)
        elif file.is_dir():
            files.extend(list_files(file, *extensions))
    return humansorted(files, alg=ns.NA | ns.G | ns.P)


def flatten_dict(content: dict[str, Any], parent_key: str = "") -> dict[str, Any]:
    items = {}
    for key, value in content.items():
        new_key = f"{parent_key}.{key}" if parent_key else key
        if isinstance(value, dict):
            items.update(flatten_dict(content=value, parent_key=new_key))
        elif isinstance(value, list) and value and isinstance(value[0], dict):
            for index, entry in enumerate(value):
                items.update(flatten_dict(content=entry, parent_key=f"{new_key}[{index}]"))
        else:
            items[new_key] = value
    return dict(humansorted(items.items(), alg=ns.NA | ns.G))


def recursive_delete(path: Path) -> None:
    for item in path.iterdir():
        if item.is_dir():
            recursive_delete(item)
        else:
            item.unlink()
    path.rmdir()


def delete_empty_folders(folder: Path) -> None:
    if folder.is_dir():
        for subfolder in folder.iterdir():
            if subfolder.is_dir():
                delete_empty_folders(subfolder)
        if not any(folder.iterdir()):
            folder.rmdir()
            LOGGER.info("Deleted empty folder: %s", folder)


def blank_is_none(value: str) -> str | None:
    """Enforces blank strings to be None."""
    return value if value else None



__all__ = [
    "BaseArchive",
    "CB7Archive",
    "CBRArchive",
    "CBTArchive",
    "CBZArchive",
    "get_archive",
    "get_archive_class",
]

from pathlib import Path
from tarfile import is_tarfile
from zipfile import is_zipfile

from rarfile import is_rarfile

from perdoo.archives._base import BaseArchive
from perdoo.archives.cb7 import CB7Archive
from perdoo.archives.cbr import CBRArchive
from perdoo.archives.cbt import CBTArchive
from perdoo.archives.cbz import CBZArchive

try:
    from py7zr import is_7zfile

    py7zr_loaded = True
except ImportError:
    py7zr_loaded = False


def get_archive(path: Path) -> BaseArchive:
    if is_zipfile(path):
        return CBZArchive(path=path)
    if is_rarfile(path):
        return CBRArchive(path=path)
    if is_tarfile(path):
        return CBTArchive(path=path)
    if py7zr_loaded and is_7zfile(path):
        return CB7Archive(path=path)
    raise NotImplementedError(f"{path.name} is an unsupported archive")


def get_archive_class(extension: str) -> type[BaseArchive]:
    archive_types = {"cbz": CBZArchive, "cbr": CBRArchive, "cbt": CBTArchive}
    if extension == "cb7" and py7zr_loaded:
        return CB7Archive
    if extension in archive_types:
        return archive_types[extension]
    raise NotImplementedError(f"{extension} is an unsupported archive")



__all__ = ["BaseService", "Comicvine", "Marvel", "Metron"]

from perdoo.services._base import BaseService
from perdoo.services.comicvine import Comicvine
from perdoo.services.marvel import Marvel
from perdoo.services.metron import Metron



__all__ = ["ZipFileRemove"]

from zipfile import ZipFile, ZipInfo


class ZipFileRemove(ZipFile):
    def remove(self, zinfo_or_arcname):
        """Remove a member from the archive."""

        if self.mode not in ('w', 'x', 'a'):
            raise ValueError("remove() requires mode 'w', 'x', or 'a'")
        if not self.fp:
            raise ValueError("Attempt to write to ZIP archive that was already closed")
        if self._writing:
            raise ValueError("Can't write to ZIP archive while an open writing handle exists")

        # Make sure we have an existing info object
        if isinstance(zinfo_or_arcname, ZipInfo):
            zinfo = zinfo_or_arcname
            # make sure zinfo exists
            if zinfo not in self.filelist:
                raise KeyError('There is no item %r in the archive' % zinfo_or_arcname)
        else:
            # get the info object
            zinfo = self.getinfo(zinfo_or_arcname)

        return self._remove_members({zinfo})

    def _remove_members(self, members, *, remove_physical=True, chunk_size=2**20):
        """Remove members in a zip file.
        All members (as zinfo) should exist in the zip; otherwise the zip file
        will erroneously end in an inconsistent state.
        """
        fp = self.fp
        entry_offset = 0
        member_seen = False

        # get a sorted filelist by header offset, in case the dir order
        # doesn't match the actual entry order
        filelist = sorted(self.filelist, key=lambda x: x.header_offset)
        for i in range(len(filelist)):
            info = filelist[i]
            is_member = info in members

            if not (member_seen or is_member):
                continue

            # get the total size of the entry
            try:
                offset = filelist[i + 1].header_offset
            except IndexError:
                offset = self.start_dir
            entry_size = offset - info.header_offset

            if is_member:
                member_seen = True
                entry_offset += entry_size

                # update caches
                self.filelist.remove(info)
                try:
                    del self.NameToInfo[info.filename]
                except KeyError:
                    pass
                continue

            # update the header and move entry data to the new position
            if remove_physical:
                old_header_offset = info.header_offset
                info.header_offset -= entry_offset
                read_size = 0
                while read_size < entry_size:
                    fp.seek(old_header_offset + read_size)
                    data = fp.read(min(entry_size - read_size, chunk_size))
                    fp.seek(info.header_offset + read_size)
                    fp.write(data)
                    fp.flush()
                    read_size += len(data)

        # Avoid missing entry if entries have a duplicated name.
        # Reverse the order as NameToInfo normally stores the last added one.
        for info in reversed(self.filelist):
            self.NameToInfo.setdefault(info.filename, info)

        # update state
        if remove_physical:
            self.start_dir -= entry_offset
        self._didModify = True

        # seek to the start of the central dir
        fp.seek(self.start_dir)



import logging
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory

from perdoo.archives import BaseArchive, get_archive_class
from perdoo.metadata import ComicInfo, MetronInfo
from perdoo.metadata.comic_info import Page
from perdoo.services import BaseService
from perdoo.settings import Service, Settings
from perdoo.utils import Search, list_files

LOGGER = logging.getLogger("perdoo")


def convert_file(entry: BaseArchive, output_format: str) -> BaseArchive | None:
    output = get_archive_class(output_format)
    if isinstance(entry, output):
        return entry
    LOGGER.debug("Converting '%s' to '%s'", entry.path.name, output_format)
    return output.convert(entry)


def clean_archive(entry: BaseArchive, settings: Settings) -> None:
    for filename in entry.list_filenames():
        if Path(filename).suffix not in settings.image_extensions:
            entry.remove_file(filename=filename)
            LOGGER.info("Removed '%s' from '%s'", filename, entry.path.name)


def _load_page_info(
    image_extensions: tuple[str, ...], entry: BaseArchive, comic_info: ComicInfo
) -> None:
    with TemporaryDirectory(prefix=f"{entry.path.stem}_") as temp_str:
        temp_folder = Path(temp_str)
        if not entry.extract_files(destination=temp_folder):
            return

        image_list = list_files(temp_folder, *image_extensions)
        pages = set()
        for index, img_file in enumerate(image_list):
            is_final_page = index == len(image_list) - 1
            page = next((x for x in comic_info.pages if x.image == index), None)
            pages.add(
                Page.from_path(file=img_file, index=index, is_final_page=is_final_page, page=page)
            )
        comic_info.pages = sorted(pages)


def sync_metadata(
    search: Search, services: dict[Service, BaseService | None], settings: Settings
) -> tuple[MetronInfo | None, ComicInfo | None]:
    for service_name in settings.services.order:
        if service := services.get(service_name):
            LOGGER.info("Searching %s for matching issue", type(service).__name__)
            metron_info, comic_info = service.fetch(search=search)
            if metron_info or comic_info:
                return metron_info, comic_info
    return None, None


def save_metadata(
    entry: BaseArchive, metadata: tuple[MetronInfo | None, ComicInfo | None], settings: Settings
) -> None:
    metron_info, comic_info = metadata
    if comic_info and settings.output.comic_info.create:
        if settings.output.comic_info.handle_pages:
            LOGGER.info("Processing ComicInfo Page data")
            _load_page_info(
                image_extensions=settings.image_extensions, entry=entry, comic_info=comic_info
            )
        else:
            comic_info.pages = []
        LOGGER.info("Writing 'ComicInfo.xml' to '%s'", entry.path.name)
        entry.write_file("ComicInfo.xml", comic_info.to_bytes().decode())
    if metron_info and settings.output.metron_info.create:
        LOGGER.info("Writing 'MetronInfo.xml' to '%s'", entry.path.name)
        entry.write_file("MetronInfo.xml", metron_info.to_bytes().decode())


def _rename_images_in_archive(
    image_extensions: tuple[str, ...], entry: BaseArchive, filename: str
) -> None:
    with TemporaryDirectory(prefix=f"{entry.path.stem}_") as temp_str:
        temp_folder = Path(temp_str)
        if not entry.extract_files(destination=temp_folder):
            return

        image_list = list_files(temp_folder, *image_extensions)
        pad_count = len(str(len(image_list)))

        for index, img_file in enumerate(image_list):
            img_filename = f"{filename}_{str(index).zfill(pad_count)}"
            if img_filename != img_file.stem:
                renamed_img = img_file.with_stem(img_filename)
                LOGGER.info("Renaming '%s' to '%s'", img_file.stem, renamed_img.stem)
                shutil.move(img_file, renamed_img)

        files = list_files(temp_folder, *image_extensions)
        files.extend([temp_folder / "ComicInfo.xml", temp_folder / "MetronInfo.xml"])
        archive_file = entry.archive_files(
            src=temp_folder, output_name=entry.path.stem, files=files
        )
        if not archive_file:
            LOGGER.critical("Unable to re-archive images")
            return

        entry.path.unlink(missing_ok=True)
        shutil.move(archive_file, entry.path)


def rename_file(
    entry: BaseArchive,
    metadata: tuple[MetronInfo | None, ComicInfo | None],
    settings: Settings,
    target: Path,
) -> None:
    metron_info, comic_info = metadata
    new_filepath = (
        metron_info.get_filename(settings=settings.output.naming)
        if metron_info
        else comic_info.get_filename(settings=settings.output.naming)
        if comic_info
        else None
    )
    if new_filepath is None:
        LOGGER.warning("Not enough information to rename '%s', skipping", entry.path.stem)
        return
    output = settings.output.folder / f"{new_filepath}.{settings.output.format}"

    if output == entry.path:
        return
    if output.exists():
        LOGGER.warning("'%s' already exists, skipping", output.relative_to(settings.output.folder))
        return
    output.parent.mkdir(parents=True, exist_ok=True)

    LOGGER.info(
        "Renaming '%s' to '%s'",
        entry.path.relative_to(target),
        output.relative_to(settings.output.folder.parent),
    )
    shutil.move(entry.path, output)
    entry.path = output

    new_filename = output.stem

    if all(
        x.startswith(new_filename)
        for x in entry.list_filenames()
        if Path(x).suffix.casefold() in settings.image_extensions
    ):
        return

    _rename_images_in_archive(
        image_extensions=settings.image_extensions, entry=entry, filename=new_filename
    )



__all__ = [
    "__version__",
    "get_cache_root",
    "get_config_root",
    "get_data_root",
    "get_state_root",
    "setup_logging",
]
__version__ = "0.5.0"

import logging
import os
from pathlib import Path

from rich.logging import RichHandler
from rich.traceback import install

from perdoo.console import CONSOLE


def get_cache_root() -> Path:
    cache_home = os.getenv("XDG_CACHE_HOME", default=str(Path.home() / ".cache"))
    folder = Path(cache_home).resolve() / "perdoo"
    folder.mkdir(exist_ok=True, parents=True)
    return folder


def get_config_root() -> Path:
    config_home = os.getenv("XDG_CONFIG_HOME", default=str(Path.home() / ".config"))
    folder = Path(config_home).resolve() / "perdoo"
    folder.mkdir(exist_ok=True, parents=True)
    return folder


def get_data_root() -> Path:
    data_home = os.getenv("XDG_DATA_HOME", default=str(Path.home() / ".local" / "share"))
    folder = Path(data_home).resolve() / "perdoo"
    folder.mkdir(exist_ok=True, parents=True)
    return folder


def get_state_root() -> Path:
    data_home = os.getenv("XDG_STATE_HOME", default=str(Path.home() / ".local" / "state"))
    folder = Path(data_home).resolve() / "perdoo"
    folder.mkdir(exist_ok=True, parents=True)
    return folder


def setup_logging(debug: bool = False) -> None:
    install(show_locals=True, max_frames=6, console=CONSOLE)

    console_handler = RichHandler(
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        omit_repeated_times=False,
        show_level=True,
        show_time=False,
        show_path=True,
        console=CONSOLE,
    )
    console_handler.setLevel(logging.DEBUG if debug else logging.INFO)
    console_handler.setFormatter(logging.Formatter("%(message)s"))
    file_handler = logging.FileHandler(filename=get_state_root() / "perdoo.log")
    file_handler.setLevel(logging.DEBUG if debug else logging.INFO)
    logging.basicConfig(
        format="[%(asctime)s] [%(levelname)-8s] {%(name)s} | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=logging.DEBUG if debug else logging.INFO,
        handlers=[console_handler, file_handler],
    )

    logging.getLogger("PIL").setLevel(logging.WARNING)



__all__ = [
    "ComicInfo",
    "Comicvine",
    "Marvel",
    "Metron",
    "MetronInfo",
    "Naming",
    "Output",
    "Service",
    "Services",
    "Settings",
]

from enum import Enum
from importlib.util import find_spec
from pathlib import Path
from typing import Annotated, Any, ClassVar, Literal

import tomli_w as tomlwriter
from pydantic import BaseModel, BeforeValidator, field_validator
from rich.panel import Panel

from perdoo import get_config_root, get_data_root
from perdoo.console import CONSOLE
from perdoo.utils import blank_is_none, flatten_dict

try:
    import tomllib as tomlreader  # Python >= 3.11
except ModuleNotFoundError:
    import tomli as tomlreader  # Python < 3.11


class SettingsModel(
    BaseModel,
    populate_by_name=True,
    str_strip_whitespace=True,
    validate_assignment=True,
    extra="ignore",
):
    pass


class ComicInfo(SettingsModel):
    create: bool = True
    handle_pages: bool = True


class MetronInfo(SettingsModel):
    create: bool = True


class Naming(SettingsModel):
    default: str = "{publisher-name}/{series-name}-v{volume}/{series-name}-v{volume}_#{number:3}"
    annual: Annotated[str | None, BeforeValidator(blank_is_none)] = (
        "{publisher-name}/{series-name}-v{volume}/{series-name}-v{volume}_Annual_#{number:2}"
    )
    digital_chapter: Annotated[str | None, BeforeValidator(blank_is_none)] = (
        "{publisher-name}/{series-name}-v{volume}/{series-name}-v{volume}_Chapter_#{number:3}"
    )
    graphic_novel: Annotated[str | None, BeforeValidator(blank_is_none)] = (
        "{publisher-name}/{series-name}-v{volume}/{series-name}-v{volume}_GN_#{number:2}"
    )
    hardcover: Annotated[str | None, BeforeValidator(blank_is_none)] = (
        "{publisher-name}/{series-name}-v{volume}/{series-name}-v{volume}_HC_#{number:2}"
    )
    limited_series: Annotated[str | None, BeforeValidator(blank_is_none)] = None
    omnibus: Annotated[str | None, BeforeValidator(blank_is_none)] = (
        "{publisher-name}/{series-name}-v{volume}/{series-name}-v{volume}_OB_#{number:2}"
    )
    one_shot: Annotated[str | None, BeforeValidator(blank_is_none)] = None
    single_issue: Annotated[str | None, BeforeValidator(blank_is_none)] = None
    trade_paperback: Annotated[str | None, BeforeValidator(blank_is_none)] = (
        "{publisher-name}/{series-name}-v{volume}/{series-name}-v{volume}_TPB_#{number:2}"
    )


class Output(SettingsModel):
    comic_info: ComicInfo = ComicInfo()
    folder: Path = get_data_root()
    format: Literal["cb7", "cbt", "cbz"] = "cbz"
    metron_info: MetronInfo = MetronInfo()
    naming: Naming = Naming()

    @field_validator("format", mode="before")
    def validate_format(cls, value: str) -> str:
        if value == "cb7" and find_spec("py7zr") is None:
            raise ImportError("Install Perdoo with the cb7 dependency group to use CB7 files.")
        return value


class Comicvine(SettingsModel):
    api_key: Annotated[str | None, BeforeValidator(blank_is_none)] = None


class Marvel(SettingsModel):
    public_key: Annotated[str | None, BeforeValidator(blank_is_none)] = None
    private_key: Annotated[str | None, BeforeValidator(blank_is_none)] = None


class Metron(SettingsModel):
    password: Annotated[str | None, BeforeValidator(blank_is_none)] = None
    username: Annotated[str | None, BeforeValidator(blank_is_none)] = None


class Service(str, Enum):
    COMICVINE = "Comicvine"
    MARVEL = "Marvel"
    METRON = "Metron"


class Services(SettingsModel):
    comicvine: Comicvine = Comicvine()
    marvel: Marvel = Marvel()
    metron: Metron = Metron()
    order: tuple[Service, ...] = (Service.METRON, Service.MARVEL, Service.COMICVINE)


def _stringify_values(content: dict[str, Any]) -> dict[str, Any]:
    output = {}
    for key, value in content.items():
        if isinstance(value, bool):
            value = str(value)
        if not value:
            continue
        if isinstance(value, dict):
            value = _stringify_values(content=value)
        elif isinstance(value, list | tuple | set):
            value = [_stringify_values(content=x) if isinstance(x, dict) else str(x) for x in value]
        else:
            value = str(value)
        output[key] = value
    return output


class Settings(SettingsModel):
    _file: ClassVar[Path] = get_config_root() / "settings.toml"

    image_extensions: tuple[str, ...] = (".jpg", ".jpeg", ".png", ".webp", ".jxl")
    output: Output = Output()
    services: Services = Services()

    @classmethod
    def load(cls) -> "Settings":
        if not cls._file.exists():
            cls().save()
        with cls._file.open("rb") as stream:
            content = tomlreader.load(stream)
        return cls(**content)

    def save(self) -> None:
        with self._file.open("wb") as stream:
            content = self.model_dump(by_alias=False, exclude_defaults=True)
            content = _stringify_values(content=content)
            tomlwriter.dump(content, stream)

    def update(self, key: str, value: Any) -> None:  # noqa: ANN401
        keys = key.split(".")
        target = self
        for entry in keys[:-1]:
            target = getattr(target, entry)
        setattr(target, keys[-1], value)

    @classmethod
    def display(cls, extras: dict[str, Any] | None = None) -> None:
        if extras is None:
            extras = {}
        default = flatten_dict(content=cls().model_dump())
        file_overrides = flatten_dict(content=cls.load().model_dump())
        default_vals = [
            f"[repr.attrib_name]{k}[/]: [repr.attrib_value]{v}[/]"
            if file_overrides[k] == v
            else f"[dim][repr.attrib_name]{k}[/]: [repr.attrib_value]{v}[/][/]"
            for k, v in default.items()
        ]
        override_vals = [
            f"[repr.attrib_name]{k}[/]: [repr.attrib_value]{v}[/]"
            for k, v in file_overrides.items()
            if default[k] != v
        ]
        extra_vals = [
            f"[repr.attrib_name]{k}[/]: [repr.attrib_value]{v}[/]" for k, v in extras.items()
        ]

        CONSOLE.print(Panel.fit("\n".join(default_vals), title="Default"))
        if override_vals:
            CONSOLE.print(Panel.fit("\n".join(override_vals), title=str(cls._file)))
        if extra_vals:
            CONSOLE.print(Panel.fit("\n".join(extra_vals), title="Extras"))



import logging
from datetime import date
from enum import Enum
from pathlib import Path
from platform import python_version
from typing import Annotated

from comicfn2dict import comicfn2dict
from typer import Argument, Context, Exit, Option, Typer

from perdoo import __version__, get_cache_root, setup_logging
from perdoo.archives import CBRArchive, get_archive
from perdoo.cli import archive_app, settings_app
from perdoo.console import CONSOLE
from perdoo.main import clean_archive, convert_file, rename_file, save_metadata, sync_metadata
from perdoo.metadata import ComicInfo, MetronInfo, get_metadata
from perdoo.metadata.metron_info import InformationSource
from perdoo.services import BaseService, Comicvine, Marvel, Metron
from perdoo.settings import Service, Services, Settings
from perdoo.utils import (
    IssueSearch,
    Search,
    SeriesSearch,
    delete_empty_folders,
    list_files,
    recursive_delete,
)

app = Typer(help="CLI tool for managing comic collections and settings.")
app.add_typer(archive_app, name="archive")
app.add_typer(settings_app, name="settings")
LOGGER = logging.getLogger("perdoo")


class SyncOption(Enum):
    FORCE = "Force"
    OUTDATED = "Outdated"
    SKIP = "Skip"

    @staticmethod
    def load(value: str) -> "SyncOption":
        for entry in SyncOption:
            if entry.value.replace(" ", "").casefold() == value.replace(" ", "").casefold():
                return entry
        raise ValueError(f"'{value}' isn't a valid SyncOption")

    def __str__(self) -> str:
        return self.value


@app.callback(invoke_without_command=True)
def common(
    ctx: Context,
    version: Annotated[
        bool | None, Option("--version", is_eager=True, help="Show the version and exit.")
    ] = None,
) -> None:
    if ctx.invoked_subcommand:
        return
    if version:
        CONSOLE.print(f"Perdoo v{__version__}")
        raise Exit


def get_services(settings: Services) -> dict[Service, BaseService]:
    output = {}
    if settings.comicvine.api_key:
        output[Service.COMICVINE] = Comicvine(settings.comicvine)
    if settings.marvel.public_key and settings.marvel.private_key:
        output[Service.MARVEL] = Marvel(settings.marvel)
    if settings.metron.username and settings.metron.password:
        output[Service.METRON] = Metron(settings.metron)
    return output


def get_search_details(
    metadata: tuple[MetronInfo | None, ComicInfo | None], fallback_title: str
) -> Search:
    metron_info, comic_info = metadata

    if metron_info and metron_info.series and metron_info.series.name:
        series_id = metron_info.series.id
        source = next(iter(x.source for x in metron_info.ids if x.primary), None)
        return Search(
            series=SeriesSearch(
                name=metron_info.series.name,
                volume=metron_info.series.volume,
                year=metron_info.series.start_year,
                comicvine=series_id if source == InformationSource.COMIC_VINE else None,
                marvel=series_id if source == InformationSource.MARVEL else None,
                metron=series_id if source == InformationSource.METRON else None,
            ),
            issue=IssueSearch(
                number=metron_info.number,
                comicvine=next(
                    iter(
                        x.value for x in metron_info.ids if x.source == InformationSource.COMIC_VINE
                    ),
                    None,
                ),
                marvel=next(
                    iter(x.value for x in metron_info.ids if x.source == InformationSource.MARVEL),
                    None,
                ),
                metron=next(
                    iter(x.value for x in metron_info.ids if x.source == InformationSource.METRON),
                    None,
                ),
            ),
        )
    if comic_info and comic_info.series:
        return Search(
            series=SeriesSearch(
                name=comic_info.series,
                volume=comic_info.volume
                if comic_info.volume and comic_info.volume < 1900
                else None,
                year=comic_info.volume if comic_info.volume and comic_info.volume > 1900 else None,
            ),
            issue=IssueSearch(number=comic_info.number),
        )
    series_name = comicfn2dict(fallback_title).get("series", fallback_title).replace("-", " ")
    return Search(series=SeriesSearch(name=series_name), issue=IssueSearch())


@app.command(name="import", help="Import comics into your collection using Perdoo.")
def run(
    target: Annotated[
        Path,
        Argument(
            exists=True, help="Import comics from the specified file/folder.", show_default=False
        ),
    ],
    skip_convert: Annotated[
        bool, Option("--skip-convert", help="Skip converting comics to the configured format.")
    ] = False,
    sync: Annotated[
        SyncOption,
        Option(
            "--sync",
            "-s",
            case_sensitive=False,
            help="Sync ComicInfo/MetronInfo with online services.",
        ),
    ] = SyncOption.OUTDATED.value,
    skip_clean: Annotated[
        bool,
        Option(
            "--skip-clean",
            help="Skip removing any files not listed in the 'image_extensions' setting.",
        ),
    ] = False,
    skip_rename: Annotated[
        bool,
        Option(
            "--skip-rename",
            help="Skip organizing and renaming comics based on their MetronInfo/ComicInfo.",
        ),
    ] = False,
    clean_cache: Annotated[
        bool,
        Option(
            "--clean",
            "-c",
            show_default=False,
            help="Clean the cache before starting the synchronization process. "
            "Removes all cached files.",
        ),
    ] = False,
    debug: Annotated[
        bool, Option("--debug", help="Enable debug mode to show extra information.")
    ] = False,
) -> None:
    setup_logging(debug=debug)
    LOGGER.info("Python v%s", python_version())
    LOGGER.info("Perdoo v%s", __version__)

    settings = Settings.load()
    settings.save()
    if debug:
        Settings.display(
            extras={
                "target": target,
                "flags.skip-convert": skip_convert,
                "flags.sync": sync,
                "flags.skip-clean": skip_clean,
                "flags.skip-rename": skip_rename,
                "flags.clean-cache": clean_cache,
            }
        )
    if clean_cache:
        LOGGER.info("Cleaning Cache")
        recursive_delete(path=get_cache_root())
    services = get_services(settings=settings.services)
    if not services and sync != SyncOption.SKIP:
        LOGGER.warning("No external services configured")
        sync = SyncOption.SKIP

    entries = []
    for file in list_files(target) if target.is_dir() else [target]:
        try:
            entries.append(get_archive(file))
        except NotImplementedError as nie:  # noqa: PERF203
            LOGGER.error("%s, Skipping", nie)  # noqa: TRY400

    for index, entry in enumerate(entries):
        CONSOLE.rule(
            f"[{index + 1}/{len(entries)}] Importing {entry.path.name}",
            align="left",
            style="subtitle",
        )
        if not skip_convert:
            with CONSOLE.status(
                f"Converting to '{settings.output.format}'", spinner="simpleDotsScrolling"
            ):
                entry = convert_file(entry, output_format=settings.output.format)
        if entry is None or isinstance(entry, CBRArchive):
            continue

        metadata = get_metadata(archive=entry, debug=debug)

        if sync != SyncOption.SKIP:
            search = get_search_details(metadata=metadata, fallback_title=entry.path.stem)
            last_modified = date(1900, 1, 1)
            if sync == SyncOption.OUTDATED:
                metron_info, _ = metadata
                if metron_info and metron_info.last_modified:
                    last_modified = metron_info.last_modified.date()
            if (date.today() - last_modified).days >= 28:
                metadata = sync_metadata(search=search, services=services, settings=settings)
            else:
                LOGGER.info("Metadata up-to-date")

        if not skip_clean:
            with CONSOLE.status("Cleaning Archive", spinner="simpleDotsScrolling"):
                clean_archive(entry=entry, settings=settings)
        save_metadata(entry=entry, metadata=metadata, settings=settings)

        if not skip_rename:
            with CONSOLE.status("Renaming based on metadata", spinner="simpleDotsScrolling"):
                rename_file(entry=entry, metadata=metadata, settings=settings, target=target.parent)

    with CONSOLE.status("Cleaning up empty folders"):
        delete_empty_folders(folder=target)


if __name__ == "__main__":
    app(prog_name="Perdoo")



__all__ = ["ComicInfo", "MetronInfo", "get_metadata"]

import logging
from typing import cast

from pydantic import ValidationError

from perdoo.archives import BaseArchive
from perdoo.metadata._base import PascalModel
from perdoo.metadata.comic_info import ComicInfo
from perdoo.metadata.metron_info import MetronInfo

LOGGER = logging.getLogger("perdoo.models")


def get_metadata(
    archive: BaseArchive, debug: bool = False
) -> tuple[MetronInfo | None, ComicInfo | None]:
    filenames = archive.list_filenames()

    def read_meta_file(cls: type[PascalModel], filename: str) -> PascalModel | None:
        if filename in filenames:
            return cls.from_bytes(content=archive.read_file(filename=filename))
        return None

    metron_info = None
    try:
        metron_info = read_meta_file(cls=MetronInfo, filename="MetronInfo.xml") or read_meta_file(
            cls=MetronInfo, filename="/MetronInfo.xml"
        )
        if metron_info:
            metron_info = cast(MetronInfo, metron_info)
    except ValidationError as ve:
        if debug:
            LOGGER.error("'%s' contains an invalid MetronInfo file: %s", archive.path.name, ve)  # noqa: TRY400
        else:
            LOGGER.error("'%s' contains an invalid MetronInfo file", archive.path.name)  # noqa: TRY400
    comic_info = None
    try:
        comic_info = read_meta_file(cls=ComicInfo, filename="ComicInfo.xml") or read_meta_file(
            cls=ComicInfo, filename="/ComicInfo.xml"
        )
        if comic_info:
            comic_info = cast(ComicInfo, comic_info)
    except ValidationError as ve:
        if debug:
            LOGGER.error("'%s' contains an invalid ComicInfo file: %s", archive.path.name, ve)  # noqa: TRY400
        else:
            LOGGER.error("'%s' contains an invalid ComicInfo file", archive.path.name)  # noqa: TRY400
    return metron_info, comic_info



__all__ = ["CBTArchive"]

import logging
import shutil
import tarfile
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

from perdoo.archives._base import BaseArchive
from perdoo.utils import list_files

LOGGER = logging.getLogger(__name__)


class CBTArchive(BaseArchive):
    def list_filenames(self) -> list[str]:
        try:
            with tarfile.open(self.path, "r") as tar:
                return tar.getnames()
        except tarfile.ReadError:
            LOGGER.exception("Unable to read %s", self.path.name)
            return []

    def read_file(self, filename: str) -> bytes:
        try:
            with tarfile.open(self.path, "r") as tar:
                return tar.extractfile(filename).read()
        except (tarfile.ReadError, KeyError):
            LOGGER.exception("Unable to read %s", self.path.name)
            return b""

    def extract_files(self, destination: Path) -> bool:
        try:
            with tarfile.open(self.path, "r") as tar:
                for member in tar.getmembers():
                    # Check for path traversal attack
                    if (
                        member.islnk()
                        or member.issym()
                        or ".." in member.name
                        or member.name.startswith("/")
                    ):
                        LOGGER.warning(
                            "Potential path traversal attack detected in %s", self.path.name
                        )
                        return False
                    tar.extract(member, path=destination)
            return True
        except tarfile.ReadError:
            LOGGER.exception("")
            return False

    @classmethod
    def archive_files(cls, src: Path, output_name: str, files: list[Path]) -> Path | None:
        output_file = src.parent / f"{output_name}.cbt"
        try:
            with tarfile.open(output_file, "w:gz") as tar:
                for file in files:
                    tar.add(file, arcname=file.name)
            return output_file
        except tarfile.CompressionError:
            LOGGER.exception("")
            return None

    @staticmethod
    def convert(old_archive: BaseArchive) -> Optional["CBTArchive"]:
        with TemporaryDirectory(prefix=f"{old_archive.path.stem}_") as temp_str:
            temp_folder = Path(temp_str)
            if not old_archive.extract_files(destination=temp_folder):
                return None
            archive_file = CBTArchive.archive_files(
                src=temp_folder, output_name=old_archive.path.stem, files=list_files(temp_folder)
            )
            if not archive_file:
                return None
            new_file = old_archive.path.with_suffix(".cbt")
            old_archive.path.unlink(missing_ok=True)
            shutil.move(archive_file, new_file)
            return CBTArchive(path=new_file)



__all__ = ["Comicvine"]

import logging
import re
from datetime import datetime

from natsort import humansorted, ns
from requests.exceptions import JSONDecodeError
from rich.prompt import Confirm, Prompt
from simyan.comicvine import Comicvine as Simyan
from simyan.exceptions import ServiceError
from simyan.schemas.issue import Issue
from simyan.schemas.volume import Volume
from simyan.sqlite_cache import SQLiteCache

from perdoo import get_cache_root
from perdoo.console import CONSOLE, create_menu
from perdoo.metadata import ComicInfo, MetronInfo
from perdoo.metadata.metron_info import InformationSource
from perdoo.services._base import BaseService
from perdoo.settings import Comicvine as ComicvineSettings
from perdoo.utils import IssueSearch, Search, SeriesSearch

LOGGER = logging.getLogger(__name__)


class Comicvine(BaseService[Volume, Issue]):
    def __init__(self, settings: ComicvineSettings):
        cache = SQLiteCache(path=get_cache_root() / "simyan.sqlite", expiry=14)
        self.session = Simyan(api_key=settings.api_key, cache=cache)

    def _search_series(self, name: str | None, volume: int | None, year: int | None) -> int | None:
        name = name or Prompt.ask("Volume Name", console=CONSOLE)
        try:
            options = sorted(
                self.session.list_volumes({"filter": f"name:{name}"}),
                key=lambda x: (
                    x.publisher.name if x.publisher and x.publisher.name else "",
                    x.name,
                    x.start_year or 0,
                ),
            )
            if year:
                options = [x for x in options if x.start_year == year]
            if not options:
                LOGGER.warning(
                    "Unable to find any Volumes with the Name and StartYear: '%s %s'", name, year
                )
            search = name
            if volume:
                search += f" v{volume}"
            if year:
                search += f" ({year})"
            index = create_menu(
                options=[
                    f"{x.id} | {x.publisher.name if x.publisher and x.publisher.name else ''}"
                    f" | {x.name} ({x.start_year})"
                    for x in options
                ],
                title="Comicvine Volume",
                subtitle=f"Searching for Volume '{search}'",
                default="None of the Above",
            )
            if index != 0:
                return options[index - 1].id
            if year:
                LOGGER.info("Searching again without the StartYear")
                return self._search_series(name=name, volume=volume, year=None)
            if Confirm.ask("Search Again", console=CONSOLE):
                return self._search_series(name=None, volume=None, year=None)
            return None
        except ServiceError:
            LOGGER.exception("")
            return None
        except JSONDecodeError:
            LOGGER.error("Unable to get response from Comicvine")  # noqa: TRY400
            return None

    def fetch_series(self, search: SeriesSearch) -> Volume | None:
        series_id = search.comicvine or self._search_series(
            name=search.name, volume=search.volume, year=search.year
        )
        if not series_id:
            return None
        try:
            series = self.session.get_volume(volume_id=series_id)
            search.comicvine = series_id
            return series
        except ServiceError:
            LOGGER.exception("")
            return None
        except JSONDecodeError:
            LOGGER.error("Unable to get response from Comicvine")  # noqa: TRY400
            return None

    def _search_issue(self, series_id: int, number: str | None) -> int | None:
        try:
            options = humansorted(
                self.session.list_issues(
                    {"filter": f"volume:{series_id},issue_number:{number}"}
                    if number
                    else {"filter": f"volume:{series_id}"}
                ),
                key=lambda x: (x.number, x.name),
                alg=ns.NA | ns.G,
            )
            if not options:
                LOGGER.warning(
                    "Unable to find any Issues with the Volume and IssueNumber: '%s %s'",
                    series_id,
                    number,
                )
            index = create_menu(
                options=[f"{x.id} | {x.number} - {x.name or ''}" for x in options],
                title="Comicvine Issue",
                subtitle=f"Searching for Issue #{number}" if number else "",
                default="None of the Above",
            )
            if index != 0:
                return options[index - 1].id
            if number:
                LOGGER.info("Searching again without the IssueNumber")
                return self._search_issue(series_id=series_id, number=None)
            return None
        except ServiceError:
            LOGGER.exception("")
            return None
        except JSONDecodeError:
            LOGGER.error("Unable to get response from Comicvine")  # noqa: TRY400
            return None

    def fetch_issue(self, series_id: int, search: IssueSearch) -> Issue | None:
        issue_id = search.comicvine or self._search_issue(series_id=series_id, number=search.number)
        if not issue_id:
            return None
        try:
            issue = self.session.get_issue(issue_id=issue_id)
            search.comicvine = issue_id
            return issue
        except ServiceError:
            LOGGER.exception("")
            return None
        except JSONDecodeError:
            LOGGER.error("Unable to get response from Comicvine")  # noqa: TRY400
            return None

    def _process_metron_info(self, series: Volume, issue: Issue) -> MetronInfo | None:
        from perdoo.metadata.metron_info import (
            Arc,
            Credit,
            Id,
            Publisher,
            Resource,
            Role,
            Series,
            Url,
        )

        def load_role(value: str) -> Role:
            try:
                return Role.load(value=value.strip())
            except ValueError:
                return Role.OTHER

        return MetronInfo(
            ids=[Id(primary=True, source=InformationSource.COMIC_VINE, value=str(issue.id))],
            publisher=Publisher(id=str(series.publisher.id), name=series.publisher.name),
            series=Series(id=str(series.id), name=series.name, start_year=series.start_year),
            collection_title=issue.name,
            number=issue.number,
            summary=issue.summary,
            cover_date=issue.cover_date,
            store_date=issue.store_date,
            arcs=[Arc(id=str(x.id), name=x.name) for x in issue.story_arcs],
            characters=[Resource[str](id=str(x.id), value=x.name) for x in issue.characters],
            teams=[Resource[str](id=str(x.id), value=x.name) for x in issue.teams],
            locations=[Resource[str](id=str(x.id), value=x.name) for x in issue.locations],
            urls=[Url(primary=True, value=issue.site_url)],
            credits=[
                Credit(
                    creator=Resource[str](id=str(x.id), value=x.name),
                    roles=[
                        Resource[Role](value=load_role(value=r.strip()))
                        for r in re.split(r"[~\r\n,]+", x.roles)
                        if r.strip()
                    ],
                )
                for x in issue.creators
            ],
            last_modified=datetime.now(),
        )

    def _process_comic_info(self, series: Volume, issue: Issue) -> ComicInfo | None:
        comic_info = ComicInfo(
            title=issue.name,
            series=series.name,
            number=issue.number,
            summary=issue.summary,
            publisher=series.publisher.name if series.publisher else None,
            web=issue.site_url,
        )

        comic_info.cover_date = issue.cover_date
        comic_info.credits = {
            x.name: [r.strip() for r in re.split(r"[~\r\n,]+", x.roles) if r.strip()]
            for x in issue.creators
        }
        comic_info.character_list = [x.name for x in issue.characters]
        comic_info.team_list = [x.name for x in issue.teams]
        comic_info.location_list = [x.name for x in issue.locations]
        comic_info.story_arc_list = [x.name for x in issue.story_arcs]

        return comic_info

    def fetch(self, search: Search) -> tuple[MetronInfo | None, ComicInfo | None]:
        if not search.series.comicvine and search.issue.comicvine:
            try:
                temp = self.session.get_issue(issue_id=search.issue.comicvine)
                search.series.comicvine = temp.volume.id
            except (ServiceError, JSONDecodeError):
                pass

        series = self.fetch_series(search=search.series)
        if not series:
            return None, None

        issue = self.fetch_issue(series_id=series.id, search=search.issue)
        if not issue:
            return None, None

        metron_info = self._process_metron_info(series=series, issue=issue)
        comic_info = self._process_comic_info(series=series, issue=issue)

        return metron_info, comic_info



__all__ = ["PascalModel", "sanitize"]

import logging
import re
from collections.abc import Callable
from pathlib import Path

from pydantic.alias_generators import to_pascal
from pydantic_xml import BaseXmlModel
from rich.panel import Panel

from perdoo.console import CONSOLE
from perdoo.utils import flatten_dict

try:
    from typing import Self  # Python >= 3.11
except ImportError:
    from typing_extensions import Self  # Python < 3.11

LOGGER = logging.getLogger(__name__)


def sanitize(value: str | None) -> str | None:
    if not value:
        return value
    value = str(value)
    value = re.sub(r"[^0-9a-zA-Z&! ]+", "", value.replace("-", " "))
    value = " ".join(value.split())
    return value.replace(" ", "-")


class PascalModel(
    BaseXmlModel,
    alias_generator=to_pascal,
    populate_by_name=True,
    str_strip_whitespace=True,
    validate_assignment=True,
    coerce_numbers_to_str=True,
    extra="ignore",
    nsmap={"xsi": "http://www.w3.org/2001/XMLSchema-instance"},
    skip_empty=True,
    search_mode="unordered",
):
    @classmethod
    def from_bytes(cls, content: bytes) -> Self:
        return cls.from_xml(content)

    def to_bytes(self) -> bytes:
        return b'<?xml version="1.0" encoding="UTF-8"?>\n' + self.to_xml(
            skip_empty=True, pretty_print=True, encoding="UTF-8"
        )

    def to_file(self, file: Path) -> None:
        file.write_bytes(self.to_bytes())

    def display(self) -> None:
        content = flatten_dict(content=self.model_dump(exclude_none=True))
        content_vals = [
            f"[repr.attrib_name]{k}[/]: [repr.attrib_value]{v}[/]" for k, v in content.items()
        ]

        CONSOLE.print(Panel.fit("\n".join(content_vals), title=type(self).__name__))

    def evaluate_pattern(self, pattern_map: dict[str, Callable[[Self], str]], pattern: str) -> str:
        def replace_match(match: re.Match) -> str:
            key = match.group("key")
            padding = match.group("padding")

            if key not in pattern_map:
                LOGGER.warning("Unknown pattern: %s", key)
                return key
            value = pattern_map[key](self)

            if padding and (isinstance(value, int) or (isinstance(value, str) and value.isdigit())):
                return f"{int(value):0{padding}}"
            return sanitize(value=value) or ""

        pattern_regex = re.compile(r"{(?P<key>[a-zA-Z-]+)(?::(?P<padding>\d+))?}")
        return pattern_regex.sub(replace_match, pattern)



__all__ = ["BaseArchive"]

from abc import ABC, abstractmethod
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional

from perdoo.utils import list_files


class BaseArchive(ABC):
    def __init__(self, path: Path):
        self.path = path

    @abstractmethod
    def list_filenames(self) -> list[str]: ...

    @abstractmethod
    def read_file(self, filename: str) -> bytes: ...

    def remove_file(self, filename: str) -> bool:
        if filename not in self.list_filenames():
            return True
        with TemporaryDirectory() as temp_str:
            temp_folder = Path(temp_str)
            if not self.extract_files(destination=temp_folder):
                return False
            (temp_folder / filename).unlink(missing_ok=True)
            return (
                self.archive_files(
                    src=temp_folder, output_name=self.path.stem, files=list_files(temp_folder)
                )
                is not None
            )

    def write_file(self, filename: str, data: str) -> bool:
        with TemporaryDirectory() as temp_str:
            temp_folder = Path(temp_str)
            if not self.extract_files(destination=temp_folder):
                return False
            (temp_folder / filename).write_text(data)
            return (
                self.archive_files(
                    src=temp_folder, output_name=self.path.stem, files=list_files(temp_folder)
                )
                is not None
            )

    @abstractmethod
    def extract_files(self, destination: Path) -> bool: ...

    @classmethod
    @abstractmethod
    def archive_files(cls, src: Path, output_name: str, files: list[Path]) -> Path | None: ...

    @staticmethod
    @abstractmethod
    def convert(old_archive: "BaseArchive") -> Optional["BaseArchive"]: ...



__all__ = ["app"]

from pathlib import Path
from typing import Annotated

from typer import Argument, Option, Typer

from perdoo.archives import get_archive
from perdoo.console import CONSOLE
from perdoo.metadata import get_metadata

app = Typer(help="Commands for inspecting and managing comic archive metadata.")


@app.command(help="View the ComicInfo/MetronInfo inside a Comic archive.")
def view(
    target: Annotated[
        Path,
        Argument(dir_okay=False, exists=True, show_default=False, help="Comic to view details of."),
    ],
    hide_comic_info: Annotated[
        bool, Option("--hide-comic-info", help="Don't show the ComicInfo details.")
    ] = False,
    hide_metron_info: Annotated[
        bool, Option("--hide-metron-info", help="Don't show the MetronInfo details.")
    ] = False,
) -> None:
    archive = get_archive(path=target)
    CONSOLE.print(f"Archive format: '{type(archive).__name__[:3]}'")
    metron_info, comic_info = get_metadata(archive=archive)
    if not hide_comic_info:
        comic_info.display()
    if not hide_metron_info:
        metron_info.display()



__all__ = ["Marvel"]

import logging
from datetime import datetime

from esak.exceptions import ApiError
from esak.schemas.comic import Comic
from esak.schemas.series import Series
from esak.session import Session as Esak
from esak.sqlite_cache import SqliteCache
from natsort import humansorted, ns
from rich.prompt import Confirm, Prompt

from perdoo import get_cache_root
from perdoo.console import CONSOLE, create_menu
from perdoo.metadata import ComicInfo, MetronInfo
from perdoo.services._base import BaseService
from perdoo.settings import Marvel as MarvelSettings
from perdoo.utils import IssueSearch, Search, SeriesSearch

LOGGER = logging.getLogger(__name__)


class Marvel(BaseService[Series, Comic]):
    def __init__(self, settings: MarvelSettings):
        cache = SqliteCache(db_name=str(get_cache_root() / "esak.sqlite"), expire=14)
        self.session = Esak(
            public_key=settings.public_key, private_key=settings.private_key, cache=cache
        )

    def _search_series(self, name: str | None, volume: int | None, year: int | None) -> int | None:
        name = name or Prompt.ask("Series Name", console=CONSOLE)
        try:
            params = {"title": name}
            if year:
                params["startYear"] = year
            options = sorted(
                self.session.series_list(params=params), key=lambda x: (x.title, x.start_year)
            )
            if not options:
                LOGGER.warning(
                    "Unable to find any Series with the Title and StartYear: '%s %s'", name, year
                )
            search = name
            if volume:
                search += f" v{volume}"
            if year:
                search += f" ({year})"
            index = create_menu(
                options=[f"{x.id} | {x.title}" for x in options],
                title="Marvel Series",
                subtitle=f"Searching for Series '{search}'",
                default="None of the Above",
            )
            if index != 0:
                return options[index - 1].id
            if year:
                LOGGER.info("Searching again without the StartYear")
                return self._search_series(name=name, volume=volume, year=None)
            if Confirm.ask("Search Again", console=CONSOLE):
                return self._search_series(name=None, volume=None, year=None)
            return None
        except ApiError:
            LOGGER.exception("")
            return None

    def fetch_series(self, search: SeriesSearch) -> Series | None:
        series_id = search.marvel or self._search_series(
            name=search.name, volume=search.volume, year=search.year
        )
        if not series_id:
            return None
        try:
            series = self.session.series(_id=series_id)
            search.marvel = series_id
            return series
        except ApiError:
            LOGGER.exception("")
            return None

    def _search_issue(self, series_id: int, number: str | None) -> int | None:
        try:
            options = humansorted(
                self.session.comics_list(
                    params={"noVariants": True, "series": series_id, "issueNumber": number}
                    if number
                    else {"noVariants": True, "series": series_id}
                ),
                key=lambda x: x.issue_number,
                alg=ns.NA | ns.G,
            )
            if not options:
                LOGGER.warning(
                    "Unable to find any Comics with the Series and IssueNumber: '%s %s'",
                    series_id,
                    number,
                )
            index = create_menu(
                options=[
                    f"{x.id} | {x.series.name} #{x.issue_number} - {x.format}" for x in options
                ],
                title="Marvel Comic",
                subtitle=f"Searching for Comic #{number}" if number else "",
                default="None of the Above",
            )
            if index != 0:
                return options[index - 1].id
            if number:
                LOGGER.info("Searching again without the IssueNumber")
                return self._search_issue(series_id=series_id, number=None)
            return None
        except ApiError:
            LOGGER.exception("")
            return None

    def fetch_issue(self, series_id: int, search: IssueSearch) -> Comic | None:
        issue_id = search.marvel or self._search_issue(series_id=series_id, number=search.number)
        if not issue_id:
            return None
        try:
            issue = self.session.comic(_id=issue_id)
            search.marvel = issue_id
            return issue
        except ApiError:
            LOGGER.exception("")
            return None

    def _process_metron_info(self, series: Series, issue: Comic) -> MetronInfo | None:
        from perdoo.metadata.metron_info import (
            GTIN,
            AgeRating,
            Arc,
            Credit,
            Format,
            Id,
            InformationSource,
            Price,
            Publisher,
            Resource,
            Role,
            Series,
            Url,
        )

        def load_format(value: str) -> Format:
            try:
                return Format.load(value=value.strip())
            except ValueError:
                return Format.SINGLE_ISSUE

        def load_age_rating(value: str) -> AgeRating:
            try:
                return AgeRating.load(value=value.strip())
            except ValueError:
                return AgeRating.UNKNOWN

        def load_role(value: str) -> Role:
            try:
                return Role.load(value=value.strip())
            except ValueError:
                return Role.OTHER

        return MetronInfo(
            ids=[Id(primary=True, source=InformationSource.MARVEL, value=str(issue.id))],
            publisher=Publisher(name="Marvel"),
            series=Series(
                id=str(series.id),
                name=series.title,
                format=load_format(value=issue.format),
                start_year=series.start_year,
            ),
            collection_title=issue.title,
            number=issue.issue_number,
            stories=[Resource[str](id=str(x.id), value=x.name) for x in issue.stories],
            summary=issue.description,
            prices=[Price(country="US", value=issue.prices.print)] if issue.prices else [],
            store_date=issue.dates.on_sale,
            page_count=issue.page_count,
            arcs=[Arc(id=str(x.id), name=x.name) for x in issue.events],
            characters=[Resource[str](id=str(x.id), value=x.name) for x in issue.characters],
            gtin=GTIN(isbn=issue.isbn, upc=issue.upc) if issue.isbn and issue.upc else None,
            age_rating=load_age_rating(value=series.rating),
            urls=[Url(primary=True, value=issue.urls.detail)],
            credits=[
                Credit(
                    creator=Resource[str](id=str(x.id), value=x.name),
                    roles=[Resource[Role](value=load_role(value=x.role))],
                )
                for x in issue.creators
            ],
            last_modified=datetime.now(),
        )

    def _process_comic_info(self, series: Series, issue: Comic) -> ComicInfo | None:
        from perdoo.metadata.comic_info import AgeRating

        def load_age_rating(value: str) -> AgeRating:
            try:
                return AgeRating.load(value=value.strip())
            except ValueError:
                return AgeRating.UNKNOWN

        comic_info = ComicInfo(
            title=issue.title,
            series=series.title,
            number=issue.issue_number,
            summary=issue.description,
            publisher="Marvel",
            web=issue.urls.detail,
            page_count=issue.page_count,
            format=issue.format,
            age_rating=load_age_rating(value=series.rating),
        )

        comic_info.credits = {x.name: x.role.strip() for x in issue.creators}
        comic_info.character_list = [x.name for x in issue.characters]
        comic_info.story_arc_list = [x.name for x in issue.events]

        return comic_info

    def fetch(self, search: Search) -> tuple[MetronInfo | None, ComicInfo | None]:
        if not search.series.marvel and search.issue.marvel:
            try:
                temp = self.session.comic(_id=search.issue.marvel)
                search.series.marvel = temp.series.id
            except ApiError:
                pass

        series = self.fetch_series(search=search.series)
        if not series:
            return None, None

        issue = self.fetch_issue(series_id=series.id, search=search.issue)
        if not issue:
            return None, None

        metron_info = self._process_metron_info(series=series, issue=issue)
        comic_info = self._process_comic_info(series=series, issue=issue)

        return metron_info, comic_info
